#include "c_battle.h"
#include "c_map.h"
#include "c_options.h"
#include <windows.h>
#include <shellapi.h>

// link with shell32.lib

#ifndef NDEBUG
char err_msg_buf[1024];
#endif

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdLine, int nCmdShow) {
	WNDCLASS wc;
	HWND hwnd;
	MSG msg;
	HACCEL haccel;

	wc.style = 0;
	wc.lpfnWndProc = WndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = DLGWINDOWEXTRA;
	wc.hInstance = hInstance;
	wc.hIcon = NULL;
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = (HBRUSH) (COLOR_BTNFACE + 1);
	wc.lpszMenuName = NULL;
	wc.lpszClassName = APP_NAME;
	RegisterClass(&wc);

	haccel = LoadAccelerators(hInstance, APP_NAME);
	hwnd = CreateDialog(hInstance, MAKEINTRESOURCE(IDD_MAIN), NULL, NULL);

	while (GetMessage(&msg, NULL, 0, 0)) {
		if (!(TranslateAccelerator(hwnd, haccel, &msg) || IsDialogMessage(hwnd, &msg))) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}
	return msg.wParam;

	UNREFERENCED_PARAMETER(nCmdShow);
	UNREFERENCED_PARAMETER(szCmdLine);
	UNREFERENCED_PARAMETER(hPrevInstance);
}

LRESULT CALLBACK WndProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	static player_data_t player_data[2];
	static game_options_t options;
	static HMENU hmenu;
	static int cell_size;

	switch (uMsg) {
		case WM_CREATE:
		{
			HRGN hrgn, hrgnwnd;
			int x = GetSystemMetrics(SM_CXSCREEN);
			int y = GetSystemMetrics(SM_CYSCREEN);

			cell_size = x / (NUM_CELLS * 5);

			hmenu = GetSubMenu(LoadMenu((HINSTANCE) GetWindowLong(hwnd, GWL_HINSTANCE),
					MAKEINTRESOURCE(IDM_MAIN)), 0);
			read_options(&options, hmenu);
			srand(GetTickCount());
			MoveWindow(hwnd,
					(x / 2) - (cell_size * NUM_CELLS) - cell_size / 2,
					(y / 2) - (cell_size * (NUM_CELLS / 2)),
					cell_size * (NUM_CELLS * 2 + 1) + 2, cell_size * NUM_CELLS + 2,
					FALSE);
			SetClassLong(hwnd, GCL_HBRBACKGROUND,
					(LONG) CreateSolidBrush(options.color[clrBg]));
			hrgn = CreateRectRgn(cell_size * NUM_CELLS + 2, 0,
					cell_size * (NUM_CELLS + 1) + 1, cell_size * NUM_CELLS + 2);
			hrgnwnd = CreateRectRgn(0, 0, cell_size * (NUM_CELLS * 2 + 1) + 2, cell_size * NUM_CELLS + 2);
			CombineRgn(hrgnwnd, hrgnwnd, hrgn, RGN_DIFF);
			SetWindowRgn(hwnd, hrgnwnd, TRUE);
			DeleteObject(hrgn);

			/* OK, let the party begin!!!!...... */
			PostMessage(hwnd, WM_COMMAND, IDM_RANDOMIZE, 0);
			return 0;
		}
		case WM_NCMOUSEMOVE:
		{
			POINT cell;

			cell.x = LOWORD(lParam);
			cell.y = HIWORD(lParam);
			ScreenToClient(hwnd, &cell);
			cell.x /= cell_size;
			cell.y /= cell_size;

			if (cell.x < NUM_CELLS && !CELL_EQUAL(cell, player_data[ENEMY].cur_cell)) {
				HDC hdc = GetDC(hwnd);
				highlight_cell(player_data[ENEMY].cur_cell, cell_size, hdc, FALSE, &options);
				highlight_cell(player_data[ENEMY].cur_cell = cell, cell_size, hdc, TRUE, &options);
				ReleaseDC(hwnd, hdc);
			}
			return 0;
		}
		case WM_NCLBUTTONDOWN:
		{
			POINT pt;

			pt.x = LOWORD(lParam);
			ScreenToClient(hwnd, &pt);
			if (pt.x <= cell_size * NUM_CELLS)
				SendMessage(hwnd, WM_KEYDOWN, VK_SPACE, 0);
			break;
		}
		case WM_NCRBUTTONDOWN:
			TrackPopupMenu(hmenu, TPM_RIGHTBUTTON,
					LOWORD(lParam), HIWORD(lParam), 0, hwnd, NULL);
			return 0;
		case WM_GETDLGCODE:
			return DLGC_WANTARROWS;
		case WM_KEYDOWN:
		{
			POINT cell = player_data[ENEMY].cur_cell;
			BOOL rep_count = GetKeyState(VK_CONTROL) < 0 ? 2 : 1;

			while (rep_count--)
				switch (wParam) {
					case VK_LEFT:
						cell.x = cell.x > 0 ? cell.x - 1 : NUM_CELLS - 1;
						break;
					case VK_RIGHT:
						cell.x = cell.x < NUM_CELLS - 1 ? cell.x + 1 : 0;
						break;
					case VK_UP:
						cell.y = cell.y > 0 ? cell.y - 1 : NUM_CELLS - 1;
						break;
					case VK_DOWN:
						cell.y = cell.y < NUM_CELLS - 1 ? cell.y + 1 : 0;
						break;
					case VK_HOME:
						cell.x = 0;
						break;
					case VK_END:
						cell.x = NUM_CELLS - 1;
						break;
					case VK_PRIOR:
						cell.y = 0;
						break;
					case VK_NEXT:
						cell.y = NUM_CELLS - 1;
						break;
					case VK_SPACE:
					{
						action_t result = action(player_data, ENEMY, hwnd, cell_size);

						if (map_all_sunk(player_data[ENEMY].map)) {
							MessageBox(hwnd, TEXT("You win!"), NULL, 0);
							return SendMessage(hwnd, WM_COMMAND, IDM_RANDOMIZE, 0);
						}

						switch (result) {
							case actHit:
								LKG_SET(player_data[ENEMY], cell);
								break;
							case actSink:
								LKG_CLEAR(player_data[ENEMY]);
								break;
						}

						if (result == actMiss) {
							do {
								player_data[ME].cur_cell = get_enemy_move(&player_data[ME],
										hwnd, options.ai_level);
								result = action(player_data, ME, hwnd, cell_size);
								switch (result) {
									case actHit:
										LKG_SET(player_data[ME], player_data[ME].cur_cell);
										break;
									case actSink:
										LKG_CLEAR(player_data[ME]);
										break;
								}
								if (map_all_sunk(player_data[ME].map)) {
									MessageBox(hwnd, TEXT("Enemy wins!"), NULL, 0);
									return SendMessage(hwnd, WM_COMMAND, IDM_RANDOMIZE, 0);
								}
							} while (result == actHit || result == actSink);
						}
						return 0;
					}
				}
			if (!CELL_EQUAL(cell, player_data[ENEMY].cur_cell)) {
				HDC hdc = GetDC(hwnd);
				highlight_cell(player_data[ENEMY].cur_cell, cell_size, hdc, FALSE, &options);
				highlight_cell(player_data[ENEMY].cur_cell = cell, cell_size, hdc, TRUE, &options);
				ReleaseDC(hwnd, hdc);
			}
			return 0;
		}
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDM_RANDOMIZE:
					reset_player_data(&player_data[ENEMY]);
					reset_player_data(&player_data[ME]);
					map_randomize(player_data[ENEMY].map);
					map_randomize(player_data[ME].map);
					InvalidateRect(hwnd, NULL, TRUE);
					return 0;
				case IDM_AI_LEVEL_1:
				case IDM_AI_LEVEL_2:
				case IDM_AI_LEVEL_3:
					CheckMenuRadioItem(hmenu, IDM_AI_LEVEL_1, IDM_AI_LEVEL_3,
							LOWORD(wParam), MF_BYCOMMAND);
					options.ai_level = LOWORD(wParam) - IDM_AI_LEVEL_1 + 1;
					return 0;
				case IDM_COLOR_BG:
				case IDM_COLOR_GRID:
				case IDM_COLOR_HIGHLIGHT:
				case IDM_COLOR_SHIP:
				case IDM_COLOR_HIT:
				case IDM_COLOR_MISS:
				{
					static COLORREF cust_colors[16];
					CHOOSECOLOR cc;

					cc.lStructSize = sizeof cc;
					cc.hwndOwner = hwnd;
					cc.hInstance = NULL;
					cc.lpCustColors = cust_colors;
					cc.rgbResult = options.color[LOWORD(wParam) - IDM_COLOR_BG];
					cc.Flags = CC_FULLOPEN | CC_RGBINIT;

					if (ChooseColor(&cc)) {
						options.color[LOWORD(wParam) - IDM_COLOR_BG] = cc.rgbResult;
						if (LOWORD(wParam) == IDM_COLOR_BG)
							DeleteObject((HGDIOBJ) SetClassLong(hwnd, GCL_HBRBACKGROUND,
									(LONG) CreateSolidBrush(cc.rgbResult)));
						InvalidateRect(hwnd, NULL, TRUE);
					}
					return 0;
				}
				case IDM_ABOUT:
					ShellAbout(hwnd, PRODUCT_NAME TEXT("#"),
							TEXT("\nbrought to you by ..::[ rihad ]::.."), NULL);
					return 0;
				case IDCANCEL:
					PostMessage(hwnd, WM_CLOSE, 0, 0);
					return 0;
			}
			break;
		case WM_PAINT:
		{
			PAINTSTRUCT ps;
			HDC hdc;
			int x, y;

			hdc = BeginPaint(hwnd, &ps);
			SelectObject(hdc, CreatePen(PS_SOLID, 1, options.color[clrGrid]));
			for (x = cell_size; x <= cell_size * (NUM_CELLS * 2 + 1); x += cell_size) {
				MoveToEx(hdc, x, 0, NULL);
				LineTo(hdc, x, cell_size * NUM_CELLS);
			}
			for (y = cell_size; y <= cell_size * NUM_CELLS; y += cell_size) {
				MoveToEx(hdc, 0, y, NULL);
				LineTo(hdc, cell_size * (NUM_CELLS * 2 + 1), y);
			}
			DeleteObject(SelectObject(hdc, GetStockObject(BLACK_PEN)));
			highlight_cell(player_data[ENEMY].cur_cell, cell_size, hdc, TRUE, &options);
			map_refresh(player_data[ME].map, cell_size, hdc, ME, options.color);
			map_refresh(player_data[ENEMY].map, cell_size, hdc, ENEMY, options.color);

			EndPaint(hwnd, &ps);
			return 0;
		}
		case WM_NCHITTEST:
			if (DefWindowProc(hwnd, WM_NCHITTEST, wParam, lParam) == HTCLIENT)
				return HTCAPTION;
			break;
		case WM_CLOSE:
			save_options(&options, hmenu);
			break;
		case WM_DESTROY:
			DeleteObject((HGDIOBJ) GetClassLong(hwnd, GCL_HBRBACKGROUND));
			PostQuitMessage(0);
			return 0;
	}
	return DefWindowProc(hwnd, uMsg, wParam, lParam);
}

void highlight_cell(POINT cell, int cell_size, HDC hdc, BOOL highlight, const game_options_t *options) {
	RECT r;
	HBRUSH hbrush;

	r.left = cell_size * cell.x + 1;
	r.top = cell_size * cell.y + 1;
	r.right = r.left + cell_size - 1;
	r.bottom = r.top + cell_size - 1;

	hbrush = CreateSolidBrush(options->color[highlight ? clrHighlight : clrBg]);
	FrameRect(hdc, &r, hbrush);
	DeleteObject(hbrush);
}

void reset_player_data(player_data_t *pdata) {
	pdata->lkg.x = pdata->lkg.y = -1;
	pdata->cur_cell.x = pdata->cur_cell.y = 0;
}

action_t action(player_data_t pdata[2], int context, HWND hwnd, int cell_size) {
	map_t cur_map = pdata[context].map;
	POINT cur_cell = pdata[context].cur_cell;
	action_t retval = actVoid;

	if (cur_map[cur_cell.y][cur_cell.x] == ctFree
			|| cur_map[cur_cell.y][cur_cell.x] == ctShip) {
		RECT r;

		r.left = cur_cell.x * cell_size;
		if (context == ME)
			r.left += (NUM_CELLS + 1) * cell_size;
		r.right = r.left + cell_size;
		r.top = cur_cell.y * cell_size;
		r.bottom = r.top + cell_size;

		if (cur_map[cur_cell.y][cur_cell.x] == ctFree) {
			cur_map[cur_cell.y][cur_cell.x] = ctMiss;
			retval = actMiss;
		} else {
			int length;

			cur_map[cur_cell.y][cur_cell.x] = ctHit;
			length = map_sink_ship(cur_map, cur_cell);
			if (!length)
				retval = actHit;
			else {
				retval = actSink;
				InflateRect(&r, cell_size * abs(length), cell_size * abs(length));
			}
		}
		InvalidateRect(hwnd, &r, TRUE);
		UpdateWindow(hwnd);
	}

	return retval;
}

POINT get_enemy_move(player_data_t *pdata, HWND hwnd, int level) {
	POINT cell = { -1, -1 };

	Sleep(500);

	if (level == 1 || !LKG_EXISTS(*pdata)) {
		do {
			cell.x = rand() % NUM_CELLS;
			cell.y = rand() % NUM_CELLS;
		} while (pdata->map[cell.y][cell.x] == ctMiss || pdata->map[cell.y][cell.x] == ctHit);
	} else {
		switch (level) {
			case 2:
			{
				/* this needs some work */
				int num_free = 0, num_hit = 0;
				int i = 0;
				POINT hit;
				/* find out the number of hit cells around us*/
				while ((i = map_cell_iterator(pdata->map, pdata->lkg, &hit, i, ctHit)) != 0)
					num_hit++;
				/* find out how many free cells are around us, not diagonally */
				while ((i = map_cell_iterator(pdata->map, pdata->lkg, &cell, i, ctFree)) != 0)
					if (i % 2 == 0)
						num_free++;
				if (!num_free || num_hit == 2 || rand() < RAND_MAX / num_free) {
					/* it's a hit */
					cell = pdata->lkg;
					while (!map_cell_iterator(pdata->map, cell, &cell, 0, ctShip))
						i = map_cell_iterator(pdata->map, cell, &cell, i, ctHit) - 1;
				}
				break;
			}
			case 3:
			{
				int i = 0;

				cell = pdata->lkg;
				while (!map_cell_iterator(pdata->map, cell, &cell, 0, ctShip))
					i = map_cell_iterator(pdata->map, cell, &cell, i, ctHit) - 1;
				break;
			}
		}
	}
	return cell;

	UNREFERENCED_PARAMETER(hwnd);
}
